/* 
    V_1_1  Version (05.01.20)
	V_1_2 Version (08.01.20)
*/
// ADD:
// v2 - invert input
// v3 - encoder velocity
// v4 - reset encoders v3.42a! 6pwm 6enc initscale
//		001- (486) FFFFFFDF

// v1_1 - rt library
// V_1_2  - RS485, many board




//#include "rtapi_ctype.h"		/* isspace() */


#include "rtapi.h"		/* RTAPI realtime OS API */
#include "rtapi_app.h"		/* RTAPI realtime module decls */
#include <rtapi_pci.h>
#include "hal.h"			/* HAL public API decls */
#include <linux/types.h>



/* module information */
//MODULE_AUTHOR("");
//MODULE_DESCRIPTION("");
//MODULE_LICENSE("GPL");

/***********************************************************************
*              				 MACROS               	       *
************************************************************************/
#define MAX_BOARD_INS  	8
#define BAR_regs	0


#define reg_enc0 0
#define reg_enc1 1
#define reg_enc2 2
#define reg_enc3 3
#define reg_enc4 4
#define reg_enc5 5

#define reg_zero 6

#define reg_in 7
#define reg_out 14

#define reg_dcont0 8
#define reg_dcont1 9
#define reg_dcont2 10
#define reg_dcont3 11
#define reg_dcont4 12
#define reg_dcont5 13

#define reg_index 15
#define control_reg 30
#define reg_WD 31

#define wd_lock_ad 29
#define info_reg 30

#define bit_en 4
#define Nenc 6
#define Nout 6

#define typeRS_PRES 0x00000200

#define rs_in_out0 22
#define rs_in1 23
#define RSinupt 48


#define VENDORID_dev   (0x2105)
#define DEV_ID         (0x5555)

#define driver_NAME "to_pci"

/***********************************************************************
*                STRUCTURES                       *
************************************************************************/

typedef struct {

	hal_u32_t	cels[32];		
	
} volatile pci_data_t;


typedef struct {
	__u8  id;
	void *mem_base;
	__u32 version;
	
	int len;
	hal_bit_t *digital_in[32];    /* ptrs for digital input pins 0 - 31 */
	hal_bit_t *digital_in_n[32];
  	hal_bit_t *digital_out[32];    /* ptrs for digital output pins 0 - 31 */
	hal_bit_t *enable_dr; 
	hal_bit_t *wd_lock;
	hal_bit_t *wd_lock_n;
	hal_bit_t *index_en[Nenc];
	hal_float_t *enccounts[Nenc];
	hal_float_t *encscale[Nenc];
	hal_float_t *encvel[Nenc];

	hal_float_t *outscale[Nout];
	hal_float_t *dcontrol[Nout];

	hal_u32_t *versionb;
	hal_u32_t *type;

	//rs485
	hal_bit_t *in_rs[RSinupt];
	hal_bit_t *in_rs_n[RSinupt];
	hal_bit_t *out_rs[32];
	hal_u32_t *ain[2];
	hal_bit_t *rs485onl;
	
	__s32 enc_lv[Nenc];

} to_pci_t;

typedef struct { 
    int		find_board; 
    int		f_alloc[MAX_BOARD_INS];
    to_pci_t	*device[MAX_BOARD_INS];
} PinCET_driver_t;


static struct
rtapi_pci_device_id PinCET_pci_tbl[] = {
    {
        .vendor = VENDORID_dev,
        .device = DEV_ID,
        .subvendor = VENDORID_dev,
        .subdevice = DEV_ID,
    },

    {0,},
};


/***********************************************************************
*               GLOBAL VARIABLES                       *
************************************************************************/


/* pointer to array of structs in shared memory, 1 per port */

/* other globals */
static int comp_id;		/* component ID */
static int num_ports;		/* number of ports configured */

static PinCET_driver_t		driver;

int WD_start[MAX_BOARD_INS];

__u8 rdctr[] = {reg_dcont0,reg_dcont1,reg_dcont2,reg_dcont3,reg_dcont4,reg_dcont5};
__u8 renci[] = {reg_enc0,reg_enc1,reg_enc2,reg_enc3,reg_enc4,reg_enc5};

/***********************************************************************
*                  LOCAL FUNCTION DECLARATIONS                         *
************************************************************************/
/* These is the functions that actually do the I/O
   everything else is just init code
*/
static void update_port(void *arg, long period);

//PCI driver misc functions
  static int PinCET_pci_probe(struct rtapi_pci_dev *dev, const struct rtapi_pci_device_id *id);
  static void PinCET_pci_remove(struct rtapi_pci_dev *dev);


static struct
rtapi_pci_driver PinCET_pci_driver = {
	.name = "hal_PinCET",
	.id_table = PinCET_pci_tbl,
	.probe = PinCET_pci_probe,
	.remove = PinCET_pci_remove,
};


/***********************************************************************
*                       INIT AND EXIT CODE                             *
************************************************************************/

int rtapi_app_main(void)
{
	static to_pci_t *device_data;
	static pci_data_t *data_pci_bus;
	char name[HAL_NAME_LEN + 1];
    	int n,i,r,j,k,enableRS ;
    	int retval=0;
    
    	n = 0;
    	num_ports = 0;

    /* STEP 1: initialise the driver */
    	comp_id = hal_init(driver_NAME);
    	if (comp_id < 0) {
    		rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: hal_init() failed\n"); 
    		return -1;
    	}



	for ( i=0; i < MAX_BOARD_INS; i++) {
		driver.f_alloc[i] = 0;
		WD_start[i] = 0;
    	}

    /* STEP 2: allocate shared memory for all board data */
	for ( i=0; i < MAX_BOARD_INS; i++) {

		driver.device[i] = hal_malloc(sizeof(to_pci_t));
		if (driver.device[i] == 0) {
    			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: hal_malloc() failed\n");
			r = -3;
			goto fail0; 
		}   
		driver.f_alloc[i] = 1;
		driver.device[i]->id = MAX_BOARD_INS;
    	}
    


/***********************************************************************
*                       PCI INIT                                       *
************************************************************************/
   
	r = rtapi_pci_register_driver(&PinCET_pci_driver);

	if (r != 0) {
          rtapi_print_msg(RTAPI_MSG_ERR,"error registering PCI driver\n");
          hal_exit(comp_id);
          return r;
        }
    
	if(num_ports > MAX_BOARD_INS)
		num_ports = MAX_BOARD_INS;

	if(num_ports == 0){
		rtapi_print_msg(RTAPI_MSG_ERR,"to_pci : board not find\n");
		r = -1;
		goto fail1;
	}

	driver.find_board = num_ports;


//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################//////////////////////////////////////////

for ( j=0; j < num_ports; j++)
{
	device_data = driver.device[j];
	data_pci_bus = device_data->mem_base;

	k = j+1;
	//RESET BOARD
	data_pci_bus->cels[control_reg] = 0x600000A0;
	data_pci_bus->cels[control_reg] = 0x60000000;

	device_data->version = data_pci_bus->cels[info_reg];

	enableRS = 0;
	if(device_data->version & typeRS_PRES)
		enableRS = 1;

//////////////////////
//////////////////////
//////////////////////

    /* Export IO pin's */

    /* export write only HAL pin's for the input bit */
    for ( i=0; i<=31;i++) {
		retval = hal_pin_bit_newf(HAL_OUT, &(device_data->digital_in[i]),comp_id, "to_pci.%d.pins.pin-%02d-in", k, i);
      	if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR, "to_pci: ERROR: port %d var export failed with err=%i\n", k + 1,retval);
			r = -1;
			goto fail1;
		}
      
      	retval = hal_pin_bit_newf(HAL_OUT, &(device_data->digital_in_n[i]),comp_id, "to_pci.%d.pins.pin-%02d-in-n", k, i);
      	if (retval < 0) {
        	rtapi_print_msg(RTAPI_MSG_ERR, "to_pci: ERROR: port %d var export failed with err=%i\n", k + 1,retval);
        	r = -1;
        	goto fail1;
      	}
      
    }

    /* export read only HAL pin's for the output bit */
    for ( i=0; i<=15;i++) {
		retval = hal_pin_bit_newf(HAL_IN, &(device_data->digital_out[i]),comp_id, "to_pci.%d.pins.pin-%02d-out", k, i);

		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", k + 1,retval);
			r = -1;
			goto fail1;
		}
    }


///////////////////////////
///////////////////////////
///////////////////////////

	/* export read only HAL pin's for the control bit */

	retval = hal_pin_bit_newf(HAL_IN, &(device_data->enable_dr),comp_id, "to_pci.%d.enable_drive", k);

	if (retval < 0) {
		rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: enable_dr var export failed with err=%i\n",retval);
		r = -1;
		goto fail1;
	}
/////////////////////////////
/////////////////////////////
/////////////////////////////

	/* export encoder signal */

	for ( i=0; i<Nenc;i++) {
		/* encoder_count */
		retval = hal_pin_float_newf(HAL_OUT, &(device_data-> enccounts[i]), comp_id, "to_pci.%d.feedback.encoder%d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR:  err encoder=%i\n", retval);
			r = -1;
			goto fail1;
		}
				/* encoder_scale */
		retval = hal_pin_float_newf(HAL_IN, &(device_data-> encscale[i]), comp_id, "to_pci.%d.feedback.enc_scale%d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR:  err enc_scale=%i\n", retval);
			r = -1;
			goto fail1;
		}
		*device_data-> encscale[i] = 1;

				/* index */
		retval = hal_pin_bit_newf(HAL_IO, &(device_data->index_en[i]),comp_id, "to_pci.%d.feedback.index_en%01d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", n + 1,retval);
			r = -1;
			goto fail1; 
		}  
			/* encoder_velocity */
		retval = hal_pin_float_newf(HAL_IN, &(device_data-> encvel[i]), comp_id, "to_pci.%d.feedback.enc_vel%d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR:  err enc_velocity=%i\n", retval);
			r = -1;
			goto fail1;
		}   
	}
	

//////////////////////////////
/////////////////////////////
/////////////////////////////
	/* export control drive PWM*/


	for ( i=0; i<Nout;i++) {
		retval = hal_pin_float_newf(HAL_IN, &(device_data-> dcontrol[i]), comp_id, "to_pci.%d.PWM.dcontrol%d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR:  err dcontrol=%i\n", retval);
			r = -1;
			goto fail1;
		}
	

		retval = hal_pin_float_newf(HAL_IN, &(device_data-> outscale[i]), comp_id, "to_pci.%d.PWM.out_scale%d", k, i);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR:  err out_scale=%i\n", retval);
			r = -1;
			goto fail1;
		}
		*device_data-> outscale[i] = 1;
	}

//////////////////////////////
//////////////////////////////* RS-485 */
/////////////////////////////
   if(enableRS){
	///#### RSinput
    	for ( i=0; i < RSinupt;i++) {
		retval = hal_pin_bit_newf(HAL_OUT, &(device_data->in_rs[i]),comp_id, "to_pci.%d.pinsRS.pin-%02d-in", k, i);
		if (retval < 0) {r = -1;goto fail1;} 
		retval = hal_pin_bit_newf(HAL_OUT, &(device_data->in_rs_n[i]),comp_id, "to_pci.%d.pinsRS.pin-%02d-in-n", k, i);
		if (retval < 0) {r = -1;goto fail1;} 
	}
	///#### RSoutput
	for ( i=0; i<=31;i++) {
		retval = hal_pin_bit_newf(HAL_IN, &(device_data->out_rs[i]),comp_id, "to_pci.%d.pinsRS.pin-%02d-out", k, i);
		if (retval < 0) {r = -1;goto fail1;}  
    	}


	/* Export analog IO pin's */
	for ( i=0; i<2;i++) {
		retval = hal_pin_u32_newf(HAL_OUT, &(device_data->ain[i]),comp_id, "to_pci.%d.analogIN.chan-%02d", k, i);
		if (retval < 0) {r = -1;goto fail1;} 
	}

	retval = hal_pin_bit_newf(HAL_IO, &(device_data->rs485onl),comp_id, "to_pci.%d.RS_485_online", k);
		if (retval < 0) {r = -1;goto fail1;} 

   }

//////////////////////////////
//////////////////////////////* OTHER SIGNALS */
/////////////////////////////


	retval = hal_pin_bit_newf(HAL_IN, &(device_data->wd_lock),comp_id, "to_pci.%d.wd_lock", k);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", n + 1,retval);
			r = -1;
			goto fail1; 
		}  
	retval = hal_pin_bit_newf(HAL_IN, &(device_data->wd_lock_n),comp_id, "to_pci.%d.wd_lock_n", k);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", n + 1,retval);
			r = -1;
			goto fail1; 
		}  
	retval = hal_pin_u32_newf(HAL_IN, &(device_data->versionb),comp_id, "to_pci.%d.version_software", k);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", n + 1,retval);
			r = -1;
			goto fail1; 
		}  
	retval = hal_pin_u32_newf(HAL_IN, &(device_data->type),comp_id, "to_pci.%d.type_board", k);
		if (retval < 0) {
			rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: port %d var export failed with err=%i\n", n + 1,retval);
			r = -1;
			goto fail1; 
		}  

}


//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////
//////////////////////////////##################################################################///////////////////////////////////////////





/////////////////////////
////UPDATE
////////////////////////
    /* STEP 4: export function */
    rtapi_snprintf(name, sizeof(name), "to_pci.1.update");
    retval = hal_export_funct(name, update_port, &driver, 1, 0,comp_id);
    if (retval < 0) {
		rtapi_print_msg(RTAPI_MSG_ERR,"to_pci: ERROR: update write funct export failed\n");
		r = -5;
		goto fail1;
    }

    rtapi_print_msg(RTAPI_MSG_INFO,"to_pci: installed driver for %d card(s)\n", num_ports);
    hal_ready(comp_id);
    return 0;




fail1:

	//for ( i=0; i < MAX_BOARD_INS; i++) {
	//	device_data = driver.device[i];
	//	rtapi_iounmap((void *)(device_data->mem_base));
	//}

fail0:
	hal_exit(comp_id);
	return r;
}




void rtapi_app_exit(void)
{	
	rtapi_pci_unregister_driver(&PinCET_pci_driver);
	hal_exit(comp_id);
	rtapi_print_msg(RTAPI_MSG_INFO,"to_pci: exit\n"); 
}




//###########################################################################################################
//
//			FUNCTIONS
//
//###########################################################################################################
int ABS_SUB(__s32 lt1, __s32 lt2)
{
	__s32 lt3;
	if(lt1<lt2)
		lt3 = lt2-lt1;
	else
		lt3 = lt1-lt2;

	if(lt3 < 5)
		return 0;
	else	
		return 1;
}
/**************************************************************
* REALTIME PORT WRITE FUNCTION                                *
**************************************************************/

void update_port(void *arg, long period){
	PinCET_driver_t *st_driver;
	pci_data_t *pci_port; 
	to_pci_t *port;

	__u32 tmp,mask,tmask,tanmask;
	__s32 ikor;
	int pin,flagok,k,i,kol,inst;

	__s32 temp1,temp2;

	__u8 temp8;

	st_driver = arg;

	tmask = 0x60000000;
	tanmask = 0xEFFFFFFF;

	kol = st_driver->find_board;
	if(kol > MAX_BOARD_INS)
		kol = MAX_BOARD_INS;

for ( i=0; i < kol; i++) {

	port = st_driver->device[i];
	inst = port->id;
	pci_port = port->mem_base;


	if(!WD_start[inst]){
		pci_port->cels[control_reg] = 0x60000080;
		pci_port->cels[control_reg] = 0x60000000;
		tmp = pci_port->cels[info_reg];
		*port->versionb = tmp & 0x000000FF;
		*port->type = (tmp >> 8) & 0x000000FF;
		WD_start[inst] = 1;
	}


// read digital inputs 	    

	tmp = pci_port->cels[reg_in];
      	mask = 0x00000001;
	for (pin=0 ; pin < 32 ; pin++) {
		*(port->digital_in[pin]) = (tmp & mask) ? 1:0 ;
		*(port->digital_in_n[pin]) = (tmp & mask) ? 0:1 ;
		mask <<= 1;
	}

// Index....
	tmp = pci_port->cels[reg_zero];
	mask = 0x01;
	for (pin=0; pin < 5; pin++) {
		if(*port->index_en[pin] && (tmp & mask)){
			*port->index_en[pin] = 0;
		}
		mask <<= 1;
     	}
///////////// read encoders /////////////////////////////////////////////

	for ( k=0; k<Nenc;k++) {
		flagok =1;
		temp1=(__s32)pci_port->cels[renci[k]];

		while (flagok)
		{
			temp2=(__s32)pci_port->cels[renci[k]];
			flagok = ABS_SUB(temp1,temp2);
			temp1 = temp2;			
		}

		*port->enccounts[k] = temp1/(*port->encscale[k]);

		temp2 = temp1 - (port->enc_lv[k]);
		port->enc_lv[k] = temp1;
		temp2 = temp2*1000;

		*port->encvel[k] = temp2/(*port->encscale[k]);
	}

///////////// wd_lock /////////////////////////////////////////////

	if((pci_port->cels[wd_lock_ad]) & 0x00000001)
	{
		*port->wd_lock = 1;
		*port->wd_lock_n = 0;
	}
	else
	{
		*port->wd_lock = 0;
		*port->wd_lock_n = 1;
	}


////////////////////ALL OUTPUTS///////////////////////////////////////////////

// write index_enable outputs 
     	tmp = 0x0;
     	mask = 0x01;
     	for (pin=0; pin < 5; pin++) {
        	if (*port->index_en[pin]) {
        		tmp |= mask;
        	}
		mask <<= 1;
     	}

	tmp |= tmask;	
	pci_port->cels[reg_index] = tmp;

// write digital outputs 
     	tmp = 0x0;
     	mask = 0x01;
     	for (pin=0; pin < 16; pin++) {
        	if (*port->digital_out[pin]) {
        		tmp |= mask;
        	}
		mask <<= 1;
     	}
	tmp |= tmask;
	pci_port->cels[reg_out] = tmp;

// write control reg  control_reg
	tmp = 0;
	if (*port->enable_dr) {
		tmp |= 0x00000010;//(1 << bit_en)
        }

	tmp |= tmask;
	tmp &= 0xFFFFFFDF;  //encoders not reset 
	pci_port->cels[control_reg] = tmp;
  
// out to drive

	for ( k=0; k<Nout;k++) {
		ikor = (*port->outscale[k])*((*(port-> dcontrol[k]))/10)*0xffff;	
		ikor |= tmask;
		ikor &= tanmask;
		pci_port->cels[rdctr[k]] = ikor;
	} 


   ////RS-485 ports

   if(port->version & typeRS_PRES)
   {
	// read digital inputs RS
	temp8 = 32;
	if(RSinupt < 32) temp8 = RSinupt;
	    
	tmp = pci_port->cels[rs_in_out0];
      	mask = 0x00000001;
	for (pin=0 ; pin < temp8 ; pin++) {
		*(port->in_rs[pin]) = (tmp & mask) ? 1:0 ;
		*(port->in_rs_n[pin]) = (tmp & mask) ? 0:1 ;
		mask <<= 1;
	}
	if(RSinupt > 32 ){
		tmp = pci_port->cels[rs_in1]; 
      		mask = 0x00000001;
		for (pin=32 ; pin < RSinupt ; pin++) {
			*(port->in_rs[pin]) = (tmp & mask) ? 1:0 ;
			*(port->in_rs_n[pin]) = (tmp & mask) ? 0:1 ;
			mask <<= 1;
		}
	}

	// read analog inputs
	__u32 avol0, avol1;
	tmp = pci_port->cels[rs_in1];
	tmp >>= 16;
	avol0 = tmp & 0x000000FF;
	//*(port->ain[0]) = tmp & 0x000000FF;
	tmp >>= 8;
	avol1 = tmp & 0x000000FF;
	//*(port->ain[1]) = tmp & 0x000000FF;

	*(port->ain[0]) = (avol0*100) >> 7;
	*(port->ain[1]) = (avol1*100) >> 7;

	////RS-485 status
	tmp = pci_port->cels[wd_lock_ad];
	if(tmp & 0x00000002)
		*port->rs485onl = 1;
	else
		*port->rs485onl = 0;

	// write digital outputs RS 
     	tmp = 0x0;
     	mask = 0x01;
     	for (pin=0; pin < 32; pin++) {
        	if (*port->out_rs[pin]) {
        		tmp |= mask;
        	}
		mask <<= 1;
     	}
	tmp |= tmask; 
	pci_port->cels[rs_in_out0] = tmp;
   }


// WD 
	pci_port->cels[reg_WD] = 0x6000000A;

}

}




//###########################################################################################################
//////////////////////////////////////////////////////////////////////////////
//                          PCI driver functions                            //
//////////////////////////////////////////////////////////////////////////////
//###########################################################################################################


static int
PinCET_pci_probe(struct rtapi_pci_dev *dev, const struct rtapi_pci_device_id *id)
{	
	to_pci_t 	*probe_device_data;
	void		*mapADDR = NULL;
	

        if (num_ports >= MAX_BOARD_INS) {
          rtapi_print_msg(RTAPI_MSG_ERR,"PinCET: skipping AnyIO board, because to many boards");
          return -3;
        }

        if (rtapi_pci_enable_device(dev)) {
          rtapi_print_msg(RTAPI_MSG_ERR,"skipping AnyIO board at %s, failed to enable PCI device\n", rtapi_pci_name(dev));
          return -4;
        }

	// Map card into memory.
	mapADDR = rtapi_pci_ioremap_bar(dev, BAR_regs);
	rtapi_print_msg(RTAPI_MSG_INFO, "PinCET: Card address : %H, Len = %d.\n", mapADDR, (int)rtapi_pci_resource_len(dev, BAR_regs));
	
	probe_device_data = driver.device[num_ports];

	// Initialize device.
	probe_device_data->mem_base = mapADDR;
	probe_device_data->id = num_ports;

	num_ports++;
	return 0;
}

static void
PinCET_pci_remove(struct rtapi_pci_dev *dev)
{
        int i;
        to_pci_t		*pDevice;

        for(i = 0; i < MAX_BOARD_INS; i++){							
          // Unmap card
		if(driver.f_alloc[i] == 1){
			pDevice = driver.device[i];
			rtapi_iounmap((void *)(pDevice->mem_base));
			driver.f_alloc[i] = 0;
		}
        }
	rtapi_pci_disable_device(dev);
        rtapi_pci_set_drvdata(dev, NULL);
}

